/**
  ******************************************************************************
  * @file    BoostConverter.cpp
  * @author  Pascal Guttmann
  * @version v1
  * @date    14 Nov 2020
  * @brief   This file contains the sourcecode for the Boost Converter.
  ******************************************************************************
  */

#include "BoostConverter.h"
#include <Arduino.h>


static volatile uint16_t BC_TargetVoltage_int = BC_VOLT2INT(BC_INPUT_VOLTAGE);


/**
  * @brief  The BC_setup function is used to initialize the Boost Converter.
  * @param  The initial target Voltage for the Converter in Volts.
  * @retval void
  */
void BC_setup(uint8_t voltage) {

	digitalWrite(BC_PWM_PIN, LOW);
	pinMode(BC_PWM_PIN, INPUT);

	analogReference(DEFAULT);
	pinMode(BC_FB_PIN, INPUT);

	digitalWrite(BC_HVDC_PIN, LOW);
	pinMode(BC_HVDC_PIN, OUTPUT);

	BC_TargetVoltage_int = BC_VOLT2INT(voltage);

	PRR &= ~(1 << PRTIM1); //Enable Timer 1 (no PowerReduction)

	// 9bit FastPWM Mode TOP 0x01FF WGM13:0 == 0b0110
	// PRESCALER 1  CS12:0 == 0b001
	// COM1A1:0 == 0b10       OC1A (ArduPin9) is cleared on Compare Match, Set at BOTTOM
	TCCR1A = ((1 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (1 << WGM11) | (0 << WGM10));
	TCCR1B = ((0 << ICNC1)  | (0 << ICES1)  | (0 << WGM13)  | (1 << WGM12)  | (0 << CS12)  | (0 << CS11)  | (1 << CS10));
	TCCR1C = ((0 << FOC1A)  | (0 << FOC1B));

	//Counter Value
	TCNT1H = 0x00;
	TCNT1L = 0x00;

	//Output Compare Reg A (init Value)
	OCR1AH = 0x00;
	OCR1AL = 0x00;

	//Timer 1 Interrupt Mask Register
	// TOIE == 1	Enable TimerOverFlow Interrupt
	TIMSK1 = ((0 << ICIE1) | (0 << OCIE1B) | (0 << OCIE1A) | (1 << TOIE1));

	//Timer 1 Interrupt Flag Register
	TIFR1 = ((0 << ICF1) | (0 << OCF1B) | (0 << OCF1A) | (0 << TOV1));

	pinMode(BC_PWM_PIN, OUTPUT);

#ifdef BC_DEBUG
	Serial.print("Boost Converter initialized. Target Voltage:\t");
	Serial.println(BC_INT2VOLT(BC_TargetVoltage_int));
#endif
}


/**
  * @brief  Reads the Output Voltage of the Boost Converter and lights Warning LED if the measured voltage is above 120V.
  * @param  void
  * @retval Returns the value of the ADC. 0...1023 == 0V...219V
  */
static int BC_getOutputRead() {
	int reading = analogRead(BC_FB_PIN);
	if (reading > BC_VOLT2INT(120)) {
		digitalWrite(BC_HVDC_PIN, HIGH);
	}
	else {
		digitalWrite(BC_HVDC_PIN, LOW);
	}
#ifdef BC_DEBUG
	Serial.print("BC_Out [V]:\t");
	Serial.println(BC_INT2VOLT(reading));
#endif
	return reading;
}


/**
  * @brief  Returns the Outputvoltage of the Boost Converter.
  * @param  void
  * @retval The Outputvoltage in Volts.
  */
uint8_t BC_readOutputVoltage(void) {
	return BC_INT2VOLT(BC_getOutputRead());
}

/**
  * @brief  Sets the Target Outputvoltage of the Boost Converter.
  * @param  The Target Voltage in Volts.
  * @retval void
  */
void BC_setTargetVoltage(uint8_t volts) {
	BC_TargetVoltage_int = floor(BC_VOLT2INT(volts));
}

/**
  * @brief  Returns the Target Outputvoltage of the Boost Converter.
  * @param  void
  * @retval The Target Voltage in Volts.
  */
uint8_t BC_readTargetVoltage(void) {
	return ceil(BC_INT2VOLT(BC_TargetVoltage_int));
}


/**
  * @brief  Sets the PWM dutycycle by writing to Timer Register. values between 0 and 511 correspond to dutycycles of 0% to 100%.
  * @param  The Value to write in the register.
  * @retval void
  */
static void BC_setPWM(uint16_t value) {//9 Bit used
#ifdef BC_DEBUG
	Serial.print("BC_PWM:\t");
	Serial.println(value);
#endif
	OCR1AH = ((value >> 8) & 0x01);
	OCR1AL = value & 0xFF;
}


/**
  * @brief  Calculates the maximum usable dutycycle for discontinuous-conduction-mode (DCM) according to the current Output Voltage.
  *			This value is not allowed to drop below 64, to ensure start up capability.
  * @param  The Output Voltage of the Boost Converter.  0...1023 == 0V...219V
  * @retval The maximum usable dutycycle to sustain DCM. 0...511 == 0%...100%
  */
static uint16_t BC_maxPWM(uint16_t analogReading) {//10-Bit val from analogRead
	uint16_t ans = 511 - 5;	//5 == 1% safety margin
	ans -= (511 * BC_INPUT_VOLTAGE*BC_R_SENSE*BC_ADC_RES / BC_R_TOTAL / BC_ADC_REF) / analogReading;
	if (ans < 64) ans = 64;
	return  ans;
}


/**
  * @brief	Calculates the P portion of the PI- Control.
  * @param  The Output Voltage of the Boost Converter.  0...1023 == 0V...219V
  * @retval P portion of PI Control.
  */
static int16_t BC_Panswer(uint16_t currentReading) {
	int16_t ans = (BC_K_P)*(BC_TargetVoltage_int - currentReading);
	return ans;
}


/**
  * @brief	Calculates the I portion of the PI- Control.
  * @param  The Output Voltage of the Boost Converter.  0...1023 == 0V...219V
  * @retval I portion of PI Control. (Limited to Interval of [0;511])
  */
static uint16_t BC_Ianswer(uint16_t currentReading) {
	static float last_answer = 0;

	float ans = last_answer + (float)BC_Panswer(currentReading)*((float)BC_UPDATETIME) / ((float)BC_T_I*1000);
	if (ans > 511) ans = 511;
	if (ans < 0) ans = 0;

	last_answer = ans;
	return ans;
}

/**
  * @brief	Calculates the PI - Control.
  * @param  The Output Voltage of the Boost Converter.  0...1023 == 0V...219V
  * @retval PI Control - PWM Value. 0...511 == 0%...100% (Limited to Interval of [0;511])
  */
static uint16_t BC_PIanswer(uint16_t currentReading) {
	uint16_t ans = max(0, BC_Panswer(currentReading)) + BC_Ianswer(currentReading);
	if (ans < 0) ans = 0;
	if (ans > 511) ans = 511;
	return ans;
}

/**
  * @brief  Sets the BoostConverter into StandBy Mode or wakes it up.
  * @param  Standby == TRUE, On == FALSE
  * @retval void
  */
void BC_setStandby(bool onoff) {
	static uint8_t oldTarget = 0;
	if (onoff) {
		oldTarget = BC_readTargetVoltage();
		BC_setTargetVoltage(0);
		pinMode(BC_PWM_PIN, INPUT);
	}
	else {
		pinMode(BC_PWM_PIN, OUTPUT);
		BC_setTargetVoltage(oldTarget);
	}
}

/**
  * @brief  This Interrupt Service Routine is called every 32us (f = 31.250 Hz) and recalculates the Output for the PWM every 16th time (f=1953.125Hz).
  * @param  void
  * @retval void
  */
ISR(TIMER1_OVF_vect) {
	//Timer 1 Interrupt Mask Register
	// TOIE == 0	Disable TimerOverFlow Interrupt
	TIMSK1 = ((0 << ICIE1) | (0 << OCIE1B) | (0 << OCIE1A) | (0 << TOIE1));

	static volatile uint8_t i = 0;
	if (i++ % 16 == 0) {
		uint16_t voltage = BC_getOutputRead();
		uint16_t pwm = min(BC_PIanswer(voltage), BC_maxPWM(voltage));
		BC_setPWM(pwm);

		//if Output stays below 35V for approx. 1 second turn off BC
		if (voltage < BC_VOLT2INT(35) && BC_readTargetVoltage() > 35) {
			static uint16_t count = 0;
			if (count++ > 2000) {
				pinMode(BC_PWM_PIN, INPUT);
				BC_setTargetVoltage(0);
				count = 0;
			}
		}
	}

	//Timer 1 Interrupt Mask Register
	// TOIE == 1	Enable TimerOverFlow Interrupt
	TIMSK1 = ((0 << ICIE1) | (0 << OCIE1B) | (0 << OCIE1A) | (1 << TOIE1));
}
