/**
  ******************************************************************************
  * @file    ClockFunctions.cpp
  * @author  Pascal Guttmann
  * @version v1
  * @date    15 Nov 2020
  * @brief   This file contains the sourcecode to display something on the Nixie Tubes.
  ******************************************************************************
  */

#include "ClockFunctions.h"
#include <Arduino.h>
#include "ShiftRegister.h"
#include <time.h>
#include <EEPROM.h>
#include "VisualUserInterface.h"

#define CF_SERVICEINTERVAL	4380 //Service Every 4380 Hours (approx. 0.5 Years)

//Static Prototypes
static void CF_EEPROM_Serialdump(void);
static uint32_t CF_getHourCounter(void);
static uint32_t CF_getServiceTimestamp(void);

/*
	EEPROM MAP:
	0x000 - 0x0FF	Service Hour Counter | 64x uint32_t
	0x3F0			Backlight Start Mode | 1x uint8_t
	0x3F4			Flicker on/off       | 1x uint8_t (0 == off)
	0x3F8 - 0x3FB	lastServiceTimestamp | 1x uint32_t
	0x3FF			Reserved for Clear EEPROM Lock (EEPROM cleared if(!42)) | 1x uint8_t
*/

/**
  * @brief	source contains the time, date, year and an  alternative source to be
  *			displayed on the Tubes.
  *			source[0][x] == Time, source[1][x] == Date,	source[2][x] == Year, source[3][x] == alt
  *			source[x][0] == Least Significant Digit,	source[x][3] == Most Significant Digit
  */
static volatile uint8_t sources[4][4] = {
	{0,0,0,0}, //Time
	{0,0,0,0}, //Date
	{0,0,0,0}, //Year
	{0,0,0,0}, //Alt				
};

/**
  * @brief	srcslct selects the source to be displayed on the Tubes.
  *			displayed on the Tubes.
  *			srcslct == 0 (Time), srcslct == 1 (Date), srcslct == 2 (Year), srcslct == 3 (Alt)
  */
static volatile uint8_t srcslct = 0;

/**
  * @brief	mask contains the status of each individual Tube. (on/off)
  *			mask[x] == 0 (Tube on), mask[x] != 0 (Tube off)
  *			mask[0] == Least Significant Digit, mask[3] == Most Significant Digit
  */
static volatile uint8_t mask[4] = { 1,1,1,1 };


static volatile uint16_t isr_iteration = 0;

static volatile bool servicePending = false;

static volatile bool CF_isStandby = false;

/**
  * @brief	Initializes Timer2 to increment the Time, Date and Year and to refresh the Tube display.
  * @param	void
  * @retval void
  */
void CF_setup(void) {
	//TIMER2 setup for OVF to refresh display and init increment of time date and year

	PRR &= ~(1 << PRTIM2); //Enable Timer 2 (no PowerReduction)

	// FastPWM Mode TOP is 0xFF		WGM22:0 == 0b011
	// PRESCALER 1024				CS12:0  == 0b111
	TCCR2A = ((0 << COM2A1) | (0 << COM2A0) | (0 << COM2B1) | (0 << COM2B0) | (1 << WGM21) | (1 << WGM20));
	TCCR2B = ((0 << FOC2A)  | (0 << FOC2B)  | (0 << WGM22)  | (1 << CS22)   | (1 << CS21)  | (1 << CS20));

	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 = ((0 << OCIE2B) | (1 << OCIE2A) | (1 << TOIE2));
	
	// Timer 2 Interrupt Flag Register
	TIFR2  = ((0 << OCF2B) | (0 << OCF2A) | (0 << TOV2));

	// Asynchronous Status Register
	ASSR = 0x00;


	CF_setTime_unix((uint32_t)0);

	//Clears EEPROM first time
	if (EEPROM.read(EEPROM.length() - 1) != 42) {
		for (int i = 0; i < EEPROM.length(); i++) {
			EEPROM.write(i, 0);
		}
		EEPROM.write(EEPROM.length() - 1, 42);
		Serial.println("EEPROM CLEARED!");
	}

	if (CF_getHourCounter() - CF_getServiceTimestamp() >= CF_SERVICEINTERVAL) {
		servicePending = true;
	}

#ifdef CF_DEBUG
	CF_EEPROM_Serialdump();
	Serial.println("Clock Function initialized.");
#endif
}

static void CF_EEPROM_Serialdump(void) {
	Serial.println("EEPROM DUMP: (Decimals)");
	for (int i = 0; i < EEPROM.length() / 8; i++) {
		Serial.print("0x");
		Serial.print(8 * i, HEX);
		Serial.print("|\t");
		for (int j = 0; j < 8; j++) {
			Serial.print(EEPROM.read(i * 8 + j));
			Serial.print("\t");
		}
		Serial.print("\n");
	}
}


/**
  * @brief	Reads the internal represantation of the Time, Date and Year and calculates 64bit Unix Time.
  * @param  void
  * @retval The corresponding Unixtime of the internal Time (Y2K Epoche)
  */
static time_t CF_getUnix(void) {
	struct tm *tm = new struct tm;
	tm->tm_sec = 0;
	tm->tm_min = 10*sources[Time][1] + 1*sources[Time][0];
	tm->tm_hour = 10 * sources[Time][3] + 1 * sources[Time][2];
	tm->tm_mday = 10 * sources[Date][3] + 1 * sources[Date][2];
	tm->tm_mon = 10 * sources[Date][1] + 1 * sources[Date][0] - 1; //jan == 0; dez == 11
	tm->tm_year = CF_YEARPRESET + 10 * sources[Year][1] + 1 * sources[Year][0] - 1900;
	tm->tm_isdst = 0;

	time_t time = mktime(tm);
	delete tm;

	return time;
}

/**
  * @brief	Overwrites internal Time represantation, according to Unix Value passed (Y2K Epoche)
  * @param  The corresponding Unixtime of the internal Time (Y2K Epoche)
  * @retval void
  */
void CF_setTime_unix(time_t time) {
	struct tm *tm = gmtime(&time); //returns pointer to internal static struct

	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	if (tm != NULL) {
		sources[Time][0] = tm->tm_min % 10;
		sources[Time][1] = tm->tm_min / 10;
		sources[Time][2] = tm->tm_hour % 10;
		sources[Time][3] = tm->tm_hour / 10;

		sources[Date][0] = (tm->tm_mon + 1) % 10;
		sources[Date][1] = (tm->tm_mon + 1) / 10;
		sources[Date][2] = tm->tm_mday % 10;
		sources[Date][3] = tm->tm_mday / 10;

		sources[Year][0] = (tm->tm_year + 1900) % 10;
		sources[Year][1] = ((tm->tm_year + 1900 - sources[Year][0]) / 10) % 10;
		sources[Year][2] = ((tm->tm_year + 1900 - sources[Year][0] - sources[Year][1]) / 100) % 10;
		sources[Year][3] = (tm->tm_year + 1900) / 1000;

		CF_startMinute();
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}


/**
  * @brief	Increments the Clock by one minute.
  * @param  void
  * @retval void
  */
void CF_increment(void) {
	CF_setTime_unix(CF_getUnix() + 60);
}


/**
  * @brief	Starts the minute. (sets seconds to zero)
  * @param  void
  * @retval void
  */
void CF_startMinute(void) {
	isr_iteration = 0;
}

/**
  * @brief	Set the Time.
  * @param  Array which contains the Time. time[0] == Least Significant Digit, time[3] == Most Significant Digit
  * @retval void
  */
void CF_setTime(uint8_t time[4]) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		sources[Time][i] = time[i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Writes the Time stored to the array specified.
  * @param  Array which is overwritten with data. Pointer needs to point to uint8_t[4] type.
  *			arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  * @retval void
  */
void CF_getTime(uint8_t* arr) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		arr[i] = sources[Time][i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Set the Date.
  * @param  Array which contains the Date. date[0] == Least Significant Digit, date[3] == Most Significant Digit
  * @retval void
  */
void CF_setDate(uint8_t date[4]) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		sources[Date][i] = date[i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Writes the Date stored to the array specified.
  * @param  Array which is overwritten with data. Pointer needs to point to uint8_t[4] type.
  *			arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  * @retval void
  */
void CF_getDate(uint8_t* arr) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		arr[i] = sources[Date][i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Set the Year.
  * @param  Array which contains the Year. year[0] == Least Significant Digit, year[1] == Most Significant Digit
  * @retval void
  */
void CF_setYear(uint8_t year[2]) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 2; i++) {
		sources[Year][i] = year[i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Writes the Year stored to the array specified.
  * @param  Array which is overwritten with data. Pointer needs to point to uint8_t[4] type.
  *			arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  * @retval void
  */
void CF_getYear(uint8_t* arr) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		arr[i] = sources[Year][i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Set the Alternative Source.
  * @param  Array which contains the Alternative data. alt[0] == Least Significant Digit, alt[3] == Most Significant Digit
  * @retval void
  */
void CF_setAlt(uint8_t alt[4]) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		sources[Alt][i] = alt[i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Writes the Alternative stored to the array specified.
  * @param  Array which is overwritten with data. Pointer needs to point to uint8_t[4] type.
  *			arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  * @retval void
  */
void CF_getAlt(uint8_t* arr) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		arr[i] = sources[Alt][i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Sets the selected source. Ignores invalid sources.
  * @param	The number of the sourc to display.
  *			0 == time, 1 == date, 2 == year, 3 == alt
  * @retval	void
  */
void CF_setSourceSelect(CF_Source src) {
	if (src <= 4) srcslct = src;
}

/**
  * @brief	Returns the selected source.
  * @param	void
  * @retval	The number of the sourc to display.
  *			0 == time, 1 == date, 2 == year, 3 == alt
  */
CF_Source CF_getSourceSelect(void) {
	return (CF_Source) srcslct;
}

/**
  * @brief	Sets the mask. _mask[i] == 0 (on), _mask[i] != 0 (off).
  * @param	The mask array. arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  * @retval	void
  */
void CF_setMask(uint8_t _mask[4]) {
	for (int i = 0; i < 4; i++) {
		mask[i] = _mask[i];
	}
}

/**
  * @brief	Writes the mask stored to the array specified.
  * @param  Array which is overwritten with data. Pointer needs to point to uint8_t[4] type.
  *			arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  * @retval void
  */
void CF_getMask(uint8_t* arr) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);
	for (int i = 0; i < 4; i++) {
		arr[i] = mask[i];
	}
	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}

/**
  * @brief	Returns an array containing the masked source for display. 0xFF representing off.
  * @param	void
  * @retval	Array which contains the masked source. arr[0] == Least Significant Digit, arr[3] == Most Significant Digit
  */
static uint8_t* CF_getMaskedSource(void) {
	static uint8_t temp[4];	//Allocate Array on HEAP not Stack, to be used outside this function
	for (int i = 0; i < 4; i++) {
		temp[i] = 0xFF;	//Default is oFF
		if (!mask[i]) temp[i] = sources[srcslct][i];
	}
	return temp;
}

/**
  * @brief  Returns the ServiceHourCounter stored on the EEPROM
  * @param  void
  * @retval The Value of the Service Hour counter. 
  */
static uint32_t CF_getHourCounter() {
	uint32_t hourCount = 0;
	uint32_t suspect = 0;
	for (int i = 0; i < 64; i++) {
		EEPROM.get(i * sizeof(hourCount), suspect);
		if (suspect > hourCount) hourCount = suspect;
	}
	return hourCount;
}


/**
  * @brief  Read the Timestamp, when the last Service was performed of ServiceHourCounter, from EEPROM.
  * @param  void
  * @retval The Value of the Service Timestamp.
  */
static uint32_t CF_getServiceTimestamp() {
	uint32_t ServiceTimestamp = 0;
	EEPROM.get(0x3F8, ServiceTimestamp);
	return ServiceTimestamp;
}

/**
  * @brief  Write a value as the Timestamp, when the Service was performed to EEPROM.
  * @param  The Value of the Service Timestamp.
  * @retval void
  */
static void CF_setServiceTimestamp(uint32_t timestamp) {
	EEPROM.put(0x3F8, timestamp);
}

/**
  * @brief  Signals the execution of the Service to the Clock. Stores the Timestamp of the Service on EEPROM.
  * @param  void
  * @retval void
  */
void CF_ServiceACK(void) {
	if (!servicePending) return;
	servicePending = false;
	CF_setServiceTimestamp(CF_getHourCounter());
	digitalWrite(VUI_SLED_R, LOW);
}

/**
  * @brief  Evaluates whether Service is pending.
  * @param  void
  * @retval True if pending, otherwise false.
  */
bool CF_service(void) {
	return servicePending;
}



/**
  * @brief  Sets the StandBy Mode of the Clock Functions.
  * @param  Stand-By == TRUE, On == FALSE
  * @retval void
  */
void CF_setStandby(bool onoff) {
	CF_isStandby = onoff;
	uint8_t arr[] = { onoff,onoff,onoff,onoff };
	CF_setMask(arr);
	SR_clear();
}

/**
  * @brief	Writes the BacklightStartMode value to the EEMPROM.
  * @param	The Value
  * @retval void
  */
void CF_setBacklightStartMode(uint8_t val) {
	EEPROM.write(0x3F0, val);
}

/**
  * @brief	Reads the BacklightStartMode value from the EEMPROM.
  * @param	void
  * @retval The Value
  */
uint8_t CF_getBacklightStartMode(void) {
	return EEPROM.read(0x3F0);
}

/**
  * @brief	Writes the FlickerMode value to the EEMPROM.
  * @param	The Value
  * @retval void
  */
void CF_setFlickerMode(bool val) {
	EEPROM.write(0x3F4, val);
}

/**
  * @brief	Reads the FlickerMode value from the EEMPROM.
  * @param	void
  * @retval The Value
  */
bool CF_getFlickerMode(void) {
	return EEPROM.read(0x3F4);
}

/**
  * @brief  This Interrupt Service Routine is called every 16.384ms (f = 61.03515625 Hz). 
  *			it updates the Display of the Tubes.
  *			every 3662 (+2 correctur 2*273ppm slower) cycles it increments the time, date and/or year. (approx. 29.87ppm to fast)
  * @param  void
  * @retval void
  */
ISR(TIMER2_OVF_vect) {
	// TimerOverFlowInt Disable		TOIE2 == 0	
	TIMSK2 &= ~(1 << TOIE2);

	if (OCR2A) {
		digitalWrite(VUI_BKLGHT, HIGH);
	}

	if (!CF_isStandby) {
		SR_printNixie(CF_getMaskedSource());
	}

	if (++isr_iteration % (3662 + 2) == 0) {
		isr_iteration = 0;
		CF_increment();
		
		static uint8_t hour = 0;
		static uint32_t hourCounter = CF_getHourCounter();
		uint32_t ServiceTimestamp = CF_getServiceTimestamp();
		
		if (++hour % 60 == 0) { //Execute approx every hour, NOT not synced with Hour of Time
			hour = 0;
			
			static uint8_t idx = 0;
			EEPROM.put(idx * sizeof(hourCounter), ++hourCounter);
			idx = (idx + 1) % 64;

			if (hourCounter - ServiceTimestamp >= CF_SERVICEINTERVAL) {
				servicePending = true;
        digitalWrite(VUI_SLED_R, HIGH);
			}	
		}
	}

	// TimerOverFlowInt Enable		TOIE2 == 1	
	TIMSK2 |= (1 << TOIE2);
}


void CF_printSrc(void) {
	String str = "";
	str += "Time: ";
	str += (char)(sources[Time][3] + '0');
	str += (char)(sources[Time][2] + '0');
	str += (char)(sources[Time][1] + '0');
	str += (char)(sources[Time][0] + '0');
	str += " Date: ";
	str += (char)(sources[Date][3] + '0');
	str += (char)(sources[Date][2] + '0');
	str += (char)(sources[Date][1] + '0');
	str += (char)(sources[Date][0] + '0');
	str += " Year: ";
	str += (char)(sources[Year][3] + '0');
	str += (char)(sources[Year][2] + '0');
	str += (char)(sources[Year][1] + '0');
	str += (char)(sources[Year][0] + '0');
	str += " Alt: ";
	str += (char)(sources[Alt][3] + '0');
	str += (char)(sources[Alt][2] + '0');
	str += (char)(sources[Alt][1] + '0');
	str += (char)(sources[Alt][0] + '0');
	Serial.print(str);
}
