/**
  ******************************************************************************
  * @file    DCF77.cpp
  * @author  Pascal Guttmann
  * @version v1
  * @date    29 Nov 2020
  * @brief   This file contains the sourcecode to receive the radio time.
  ******************************************************************************
  */

#include "DCF77.h"
#include <Arduino.h>
#include "ClockFunctions.h"
#include <time.h>
#include "VisualUserInterface.h"


#define DCF77_TCO 2
#define DCF77_PON 4 //POWER ON (low active)

#define DCF77_SPLITTIME 150
#define DCF77_RISINGJITTER 100 //in ms
#define DCF77_FALLINGJITTER 50 //in ms

#define DCF77_RXBLINK_OK	VUI_BKLGHT
#define DCF77_RXBLINK_FLT	VUI_SLED_R

typedef enum {
	invalid,
	falling,
	rising
}DCF77_Edge;


volatile uint64_t telegram = 0;
volatile uint8_t bit_no = 0;

volatile time_t unixtime = 0;

volatile bool telegramReadyFlag = false;
volatile bool DCF77_TelegramStart = false;
volatile bool DCF77_UnixTimeValid = false;

bool successful_receive = false;
bool RX_blink = false;

/**
  * @brief	Initializes the DCF77 Receive Module.
  * @param	void
  * @retval void
  */
void DCF77_setup(void) {
	DCF77_setOnOff(0);
	pinMode(DCF77_PON, OUTPUT);

	pinMode(DCF77_TCO, INPUT_PULLUP);
	attachInterrupt(digitalPinToInterrupt(DCF77_TCO), ISR_DCF77_TCO, CHANGE);

#ifdef DCF77_DEBUG
	Serial.println("DCF77 initialized.");
#endif 
}

/**
  * @brief	Set the DCF77 Receive Module On/Off.
  * @param	true = on, false = off
  * @retval void
  */
void DCF77_setOnOff(bool on) {
	digitalWrite(DCF77_PON, !on);
	DCF77_TelegramStart = false;
	DCF77_UnixTimeValid = false;
}

/**
  * @brief	Returns if since the last call of this Function there was a Successfull Receive.
  * @param	void
  * @retval success true/false
  */
bool DCF77_getSuccessfullRX(void) {
	bool temp = successful_receive;
	successful_receive = false;
	return temp;
}

/**
  * @brief	Returns the Bit in the telegram
  * @param	the bit number
  * @retval The boolean value of the bit.
  */
static bool getTelBit(uint8_t bit) {
	return (telegram >> bit) & 1ULL;
}

/**
  * @brief	Sets the Bit in the telegram, to the specified value.
  * @param	the bit number, the value to set the bit to
  * @retval void
  */
static void setTelBit(uint8_t bit, bool val) {
	if (getTelBit(bit) != val) telegram ^= (1ULL << bit_no);
}


/**
  * @brief	Reset the DCF77 Telegram.
  * @param	void
  * @retval void
  */
static void DCF77_resetBitNo(void) {
	bit_no = 0;
#ifdef DCF77_DEBUG
	Serial.println("DCF77 Telegram Reset.");
#endif 
}

/**
  * @brief	Increments the DCF77 Telegram by one bit.
  * @param	void
  * @retval void
  */
static void DCF77_incBitNo(void) {
	bit_no = (bit_no + 1) % 59;
}

/**
  * @brief	Appends the given value to the Telegram buffer. Sets the TelegramReadyFlag for the Code Routine if, the buffer overflows.
  * @param	The value to append.
  * @retval void
  */
static void DCF77_appendBit(bool val) {
	setTelBit(bit_no, val);

#ifdef DCF77_DEBUG
	Serial.print("DCF77 Bit:\t");
	Serial.print(bit_no);
#endif 

	DCF77_incBitNo();
	telegramReadyFlag = !bit_no;
	
#ifdef DCF77_DEBUG
	Serial.print("\tRX:\t");
	Serial.println(val);
	if (telegramReadyFlag) {
		Serial.println("DCF77 Telegram Ready.");
	}
#endif 
}


/**
  * @brief	Checks the Telegram in the buffer for its parity in the specified range.
  * @param	The start and the end Bit of the Sequence.
  * @retval returns true when test is passed.
  */
static bool DCF77_parityCheckBits(uint8_t start, uint8_t end) {
	uint8_t parity = 0;
	for (int i = start; i <= end; i++) {
		parity += getTelBit(i);
	}
	return !(parity % 2);
}

/**
  * @brief	Checks the Telegram in the buffer for its parity.
  * @param	void
  * @retval returns true when test is passed.
  */
static bool DCF77_parityCheckTelegram(void) {
	bool min_par = DCF77_parityCheckBits(21, 28);
	bool hour_par = DCF77_parityCheckBits(29, 35);
	bool date_par = DCF77_parityCheckBits(36, 58);
	return ( min_par && hour_par && date_par && getTelBit(20) && !getTelBit(0) );
}


/**
  * @brief	Code Routine of DCF77 to evaluate the Telegram and set Time and Date.
  * @param	void
  * @retval void
  */
void DCF77_CodeRoutine(void) {
	static uint32_t unixtime_timestamp = 0;

	if (telegramReadyFlag) {
		telegramReadyFlag = false;
#ifdef DCF77_DEBUG
		Serial.print("DCF77 Telegram arrived in Code Routine:\t");
		for (int i = 0; i < 64; i++) {
			Serial.print( getTelBit(63-i) );
		}
		Serial.print("\n");
#endif
		
		if (DCF77_parityCheckTelegram()) {
		#ifdef DCF77_DEBUG
			Serial.println("DCF77 Telegram passed Parity Checks");
		#endif
			bool invalidData = false;
			if (getTelBit(18) == getTelBit(17) || getTelBit(19)) invalidData = true;
			
			struct tm *tm = new struct tm;
			tm->tm_sec  = 0;
			tm->tm_min  = 1 * getTelBit((21)) + 2 * getTelBit((22)) + 4 * getTelBit((23)) + 8 * getTelBit((24)) + 10 * getTelBit((25)) + 20 * getTelBit((26)) + 40 * getTelBit((27));
			tm->tm_hour = 1 * getTelBit((29)) + 2 * getTelBit((30)) + 4 * getTelBit((31)) + 8 * getTelBit((32)) + 10 * getTelBit((33)) + 20 * getTelBit((34));
			tm->tm_mday = 1 * getTelBit((36)) + 2 * getTelBit((37)) + 4 * getTelBit((38)) + 8 * getTelBit((39)) + 10 * getTelBit((40)) + 20 * getTelBit((41));
			tm->tm_mon  = 1 * getTelBit((45)) + 2 * getTelBit((46)) + 4 * getTelBit((47)) + 8 * getTelBit((48)) + 10 * getTelBit((49)) - 1; //jan == 0; dez == 11
			tm->tm_year  = CF_YEARPRESET - 1900 + 1 * getTelBit((50)) + 2 * getTelBit((51)) + 4 * getTelBit((52)) + 8 * getTelBit((53)) + 10 * getTelBit((54)) + 20 * getTelBit((55)) + 40 * getTelBit((56)) + 80 * getTelBit((57));
			tm->tm_isdst = 0;

			time_t time = mktime(tm);

			//IF DCF77 Transmitts in MESZ subtract one hour, to get MEZ. (MEZ used throughout the Clock)
			///if (getTelBit(18) && !getTelBit(17)) time -= 3600;
			//Set Offset to received time, to show correct Time in Timezone.
			time += CF_TIMEZONE * 3600;


			delete tm;

			if (!invalidData) {
				unixtime = time;
				DCF77_UnixTimeValid = true;
				unixtime_timestamp = millis();
				DCF77_TelegramStart = false;
			}
		}
	}

	if (DCF77_UnixTimeValid && DCF77_TelegramStart) {
		CF_setTime_unix(unixtime);
		successful_receive = true;
		DCF77_UnixTimeValid = false;
		DCF77_TelegramStart = false;
#ifdef DCF77_DEBUG
		Serial.print("DCF77 Setting Y2K Time:\t");
		Serial.println(unixtime);
		CF_printSrc();
		Serial.print("\n");
#endif
	}

	//Invalidate Unixtime, when not used for 2,5 sec after validation
	if (millis() - unixtime_timestamp >= 2500) DCF77_UnixTimeValid = false;
}



/**
  * @brief	Detect Edge of Signal and validate against last State. If the same state is read twice, the edge is invalid. Resets the Telgram Counter to zero if invalid edge is detected.
  * @param	The current State of the TCO Pin.
  * @retval The DCF77_Edge of the TCO Pin.
  */
static DCF77_Edge DCF77_edgeDetect(bool pinState, uint32_t timestamp) {
	static bool lastPinState = LOW;
	static uint32_t lastTimestamp_rising = 0;

	if (pinState == lastPinState) {
		return invalid;
	}
	lastPinState = pinState;

	DCF77_Edge edge = (DCF77_Edge)(1 + pinState);

#ifdef DCF77_DEBUG
	//Serial.print("Internal Edge:\t");
	//Serial.println(edge);
	//Serial.print("TimeDiff:\t");
	//Serial.println(timestamp - lastTimestamp_rising);
#endif

	if (edge == rising) {

		//Rising Edge: check timing between 980...1020ms, or restart telegram
		if (!(1000 - DCF77_RISINGJITTER < timestamp - lastTimestamp_rising && timestamp - lastTimestamp_rising < 1000 + DCF77_RISINGJITTER)) {
			DCF77_resetBitNo();
			DCF77_TelegramStart = true;
		}
		lastTimestamp_rising = timestamp;
		return rising;
	}
	else { //edge == falling

		//Fallin edge: check timing between 90..110ms || 190..210ms
		if ((100 - DCF77_FALLINGJITTER < timestamp - lastTimestamp_rising && timestamp - lastTimestamp_rising < 100 + DCF77_FALLINGJITTER) ||
			(200 - DCF77_FALLINGJITTER < timestamp - lastTimestamp_rising && timestamp - lastTimestamp_rising < 200 + DCF77_FALLINGJITTER)) {
			return falling;
		}
		else {
			DCF77_resetBitNo();
			return invalid;
		}
	}
}

/**
  * @brief	Turns on off Blinking signals when Receiving Telegram
  * @param	On == true, off == false
  * @retval void
  */
void DCF77_setRXBlink(bool onoff) {
	RX_blink = onoff;
	if (onoff) {
		//Stop PWM dimming
		TIMSK2 &= ~(1 << OCIE2A);
		//set 0 so TIMER2_OVF_vect does not light LED
		OCR2A = 0;
	}
	else {
		//Start PWM dimming
		TIMSK2 |= (1 << OCIE2A);
	}
	digitalWrite(DCF77_RXBLINK_FLT, LOW);
	digitalWrite(DCF77_RXBLINK_OK, LOW);
}

/**
  * @brief	ISR at Pinchange of Pin 2. Detects long and short modulation of DCF77 Signal. Sends single bits to Telegram.
  * @param	void
  * @retval void
  */
void ISR_DCF77_TCO() {
	EIMSK &= ~(1 << INT0);

	
	bool pinState = digitalRead(DCF77_TCO);
	uint32_t timestamp = millis();
	DCF77_Edge edge = DCF77_edgeDetect(pinState, timestamp);
	static uint32_t lastTimeStamp = 0;

#ifdef DCF77_DEBUG
	//Serial.print("DCF77 Int.\t@");
	//Serial.print(timestamp);
	//Serial.print("\t IN:\t");
	//Serial.print(pinState);
	//Serial.print("\t Edge:\t");
	//Serial.println(edge);
#endif

	switch (edge) {
	case falling:
		if (RX_blink) {
			digitalWrite(DCF77_RXBLINK_FLT, LOW);
			digitalWrite(DCF77_RXBLINK_OK, LOW);
		}
		if (timestamp - lastTimeStamp < DCF77_SPLITTIME) {
			DCF77_appendBit(0);
		}
		else {
			DCF77_appendBit(1);
		}
		break;
	case rising:
		if (RX_blink) {
			digitalWrite(DCF77_RXBLINK_FLT, LOW);
			digitalWrite(DCF77_RXBLINK_OK, HIGH);
		}
		lastTimeStamp = timestamp;
		break;
	default:
		if (RX_blink) {
			digitalWrite(DCF77_RXBLINK_FLT, HIGH);
		}
		break;
	}
	EIMSK |= (1 << INT0);
}
