/**
  ******************************************************************************
  * @file    NixieClock2.ino
  * @author  Pascal Guttmann
  * @version v1
  * @date    14 Nov 2020
  * @brief   This file contains the main program for the Nixie Clock.
  ******************************************************************************
  */

#include "DCF77.h"
#include <Arduino.h>
#include "VisualUserInterface.h"
#include "BoostConverter.h"
#include "ShiftRegister.h"
#include "ClockFunctions.h"
#include "EEPROM.h"
#include "DCF77.h"


//Uncomment for Debug
///#define MAIN_DEBUG

bool isStandby = false;
bool userStandby = false;
uint32_t standby_timestamp = 0;
bool flicker = true;



void setup() {
	cli();

	Serial.begin(115200);
#ifdef MAIN_DEBUG
	Serial.println("Start up...");
#endif

	SR_setup();
	BC_setup(160);
	CF_setup();
	VUI_setup();
	DCF77_setup();

	setStandby(true);
	sei();

  bool manualClockSet = false;
	DCF77_setRXBlink(true);
	while (true) {
		DCF77_CodeRoutine();
		if (DCF77_getSuccessfullRX()) break;
    
    if (VUI_svcbtn() == Released && !svc_btn_timeout()) {
      manualClockSet = true;
      break;
    }
	}
	DCF77_setRXBlink(false);
	setStandby(false);

  if(manualClockSet){
    manClockSet();
  }

	uint8_t arr[] = { 0,0,0,0 };
	CF_setMask(arr);
	CF_setSourceSelect(Time);

	set_backlightLevel(CF_getBacklightStartMode());

  if(CF_service()){
    digitalWrite(VUI_SLED_R, HIGH);
  }
}


void loop() {
	bool timeCompActionAllowed = timeCompChange();
	
	//Short Button Press Action
	if (VUI_btn_short()) {
		if (getStandby()) { //Stand By WAKEUP
			setStandby(false);
			CF_setSourceSelect(Time);
		}
		else { 
			if (VUI_svcbtn() == Hold) { //Store Backlight Level to EEPROM
				CF_setBacklightStartMode(get_backlightLevel());
				digitalWrite(VUI_SLED_G, HIGH);
			}
			else { //Normal Button Action
				CF_Source srcslct = CF_getSourceSelect();
				if (srcslct != Alt) {
					srcslct = (CF_Source)((srcslct + 1) % 3);
					CF_setSourceSelect(srcslct);
				}
				else {
					CF_setSourceSelect(Time);
				}
			}
		}
		digitalWrite(VUI_SLED_B, LOW);
	}

	//Long Button Press Action
	if (VUI_btn_long()) {
		if (getStandby()) { //Stand By WAKEUP
			setStandby(false);
			CF_setSourceSelect(Time);
		}
		else { 
			if (VUI_svcbtn() == Hold) { //Enable Disable Flicker
				CF_setFlickerMode(!CF_getFlickerMode());
				digitalWrite(VUI_SLED_G, HIGH);
			}
			else { //Normal Button Action
				inc_backlightLevel();
			}
		}
	}

	//Service Button Press Action
	if (VUI_svcbtn() == Released && !svc_btn_timeout()) {
		if (!digitalRead(VUI_SLED_G)) { //Normal Action, only when not "Combi Press"
			CF_ServiceACK();
      
      userStandby = true;
			setStandby(!getStandby());
      

			digitalWrite(VUI_SLED_B, LOW);
		}
		digitalWrite(VUI_SLED_G, LOW);
	}

	//Execute scheduled events
	if (timeCompActionAllowed) {
		timeCompActionAllowed = false;
		run_schedule();
	}
	
	//Flicker Tubes: One Sequence then all tubes on.
	if (flicker && CF_getFlickerMode()) {
		if (flicker_tubes()) {
			flicker = false;
			uint8_t arr[] = { 0,0,0,0 };
			CF_setMask(arr);
		}
	}
	else {
		flicker = false;
	}
	Change_AltSource();
	standBy_timeout();
	source_timeout();
	DCF77_CodeRoutine();
	
}

/**
  * @brief  Routine to manually set the Clock if no DCF77 RX.
  * @param  void
  * @retval void
  */
void manClockSet(void){
  uint8_t digit = 0;
  CF_setSourceSelect(Time);

  //Reset pending BTN Presses
  VUI_btn_short();
  VUI_btn_long();
  VUI_svcbtn();
  
  while(CF_getSourceSelect() != Alt){
    
    if (VUI_btn_short() || VUI_btn_long()) {
      if (VUI_svcbtn() == Hold) {
       CF_setSourceSelect(CF_getSourceSelect() + 1);
       digitalWrite(VUI_SLED_G, HIGH);
      }
      else{
        //inc value of digit
        uint8_t arr[] = { 0,0,0,0 };
        switch(CF_getSourceSelect()) {
          case Time:
            CF_getTime(&arr[0]);
            break;
          case Date:
            CF_getDate(&arr[0]);
            break;
          case Year:
            CF_getYear(&arr[0]);
            break;
          default:
            break;
        }
        arr[digit] = (arr[digit]+1) % 10;
        switch(CF_getSourceSelect()) {
          case Time:
            CF_setTime(arr);
            break;
          case Date:
            CF_setDate(arr);
            break;
          case Year:
            uint8_t sarr[] = { arr[0], arr[1] };
            CF_setYear(sarr);
            break;
          default:
            break;
        }
      }
    }
    
    if (VUI_svcbtn() == Released && !svc_btn_timeout()) {
      if(digitalRead(VUI_SLED_G)){
        digit = 0;
      }
      else{
        digit = (digit + 1) % 4;
      }
      digitalWrite(VUI_SLED_G, LOW);
    }
    
    blinkDigit(digit);
    //Blink active digit
  }

  //Light all Tubes
  uint8_t arr[] = { 0,0,0,0 };
  CF_setMask(arr);
}

void blinkDigit(uint8_t digit){
  static uint32_t timestamp = 0;
  if(millis() - timestamp > 350){
    timestamp = millis();
    uint8_t arr[] = { 0,0,0,0 };
    CF_getMask(&arr[0]);
    for(int i = 0; i < 4; i++){
      arr[i] = (i == digit) && (arr[digit] == 0);
    }
    CF_setMask(arr);
    
  }
}


/**
  * @brief  Executes scheduled events.
  * @param  void
  * @retval void
  */
void run_schedule(void) {
	//Enter Stand By to Receive Time
	if (timeComp(0x0200)) {
		setStandby(true);
		//Call to clear SuccessfulRX
		DCF77_getSuccessfullRX();
	}

	//Wake Up and start to care tubes
	if (timeComp(0x0300)) {
    if(!userStandby){
		  setStandby(false);
    }
		CF_setSourceSelect(Alt);

    if (!DCF77_getSuccessfullRX()) {
      if(!userStandby){
  		  digitalWrite(VUI_SLED_B, HIGH);
      }
  	}
	}

	//Stop to care tubes
	if (timeComp(0x0500)) {
		CF_setSourceSelect(Time);
	}
}




/**
  * @brief  Cycles the Backlight through its dimm states.
  * @param  void
  * @retval void
  */
void inc_backlightLevel(void) {
	set_backlightLevel((get_backlightLevel() + 1) % 4);
}


/**
  * @brief  Returns the current backlight level.
  * @param  void
  * @retval Backlight Level 0...3
  */
uint8_t get_backlightLevel(void) {
	if (OCR2A == 64) return 1;
	if (OCR2A == 127) return 2;
	if (OCR2A == 255) return 3;
	return 0;
}

/**
  * @brief  Sets the Backlight level.
  * @param  The Level 0...3
  * @retval void
  */
void set_backlightLevel(uint8_t level) {
	switch (level) {
	case 1:
		set_backlight(64);
		break;
	case 2:
		set_backlight(127);
		break;
	case 3:
		set_backlight(255);
		break;
	case 0:
	default:
		set_backlight(0);
		break;
	}
}

/**
  * @brief  Sets the duty cycle of the PWM for the Backlight.
  * @param  The PWM Value of the Backlight.
  * @retval void
  */
void set_backlight(uint8_t value) {
	OCR2A = value;
}

/**
  * @brief  Turns off the Backlight.
  * @param  void
  * @retval void
  */
ISR(TIMER2_COMPA_vect) {
	digitalWrite(VUI_BKLGHT, LOW);
}

/**
  * @brief  Turns the Stand-By Mode of the Clock On/Off.
  * @param  On == TRUE, Off == FALSE
  * @retval void
  */
void setStandby(bool onoff) {
#ifdef MAIN_DEBUG
	Serial.print("Stand By\t");
	Serial.print(onoff);
	Serial.println();
#endif
	if (onoff == getStandby()) return;
	static uint8_t bklght_save = 0;
	isStandby = onoff;
	BC_setStandby(onoff);
	if (!onoff) { //WAKE UP
    userStandby = false;
    while (BC_readOutputVoltage() < 0.98*BC_readTargetVoltage()) {}
		flicker = true;

		//Recover Backlight
		OCR2A = bklght_save;
	}
	else { //SLEEP
		standby_timestamp = millis();

		//set 0 so TIMER2_OVF_vect does not light LED
		bklght_save = OCR2A;
		OCR2A = 0;
	}
	CF_setStandby(onoff);
	DCF77_setOnOff(onoff);
}

/**
  * @brief  Returns true if the time stored is equal to the time specified.
  * @param  Time in "hex Format" (0x1634 == 16:34Uhr)
  * @retval Equal == TRUE, Unequal == FALSE
  */
bool timeComp(uint16_t time) {
	uint8_t time_arr[4];
	CF_getTime(&time_arr[0]);
	for (int i = 0; i < 4; i++) {
		if (time_arr[i] != ((time >> 4 * i) & 0x000F)) {
			return false;
		}
	}
	return true;
}

/**
  * @brief  Returns true if the stored Time has changed since last call to this function.
  * @param  void
  * @retval Change == TRUE, Unchanged == FALSE
  */
bool timeCompChange(void) {
	static uint8_t last_time_arr[] = { 0xFF,0xFF,0xFF,0xFF };
	uint8_t time_arr[4];
	CF_getTime(&time_arr[0]);

	bool change = false;

	for (int i = 0; i < 4; i++) {
		if (time_arr[i] != last_time_arr[i]){
			change = true;
			last_time_arr[i] = time_arr[i];
		}
	}
	return change;
}


/**
  * @brief  Returns the Stand-By Mode of the Clock.
  * @param  void
  * @retval On == TRUE, Off == FALSE
  */
bool getStandby(void) {
	return isStandby;
}

/**
  * @brief  Automatically disables Standby after period of time.
  * @param  void
  * @retval void
  */
void standBy_timeout(void) {
	if (millis() - standby_timestamp > 6*3600000ULL && getStandby()) {
		setStandby(false);
	}
}

/**
  * @brief  Automatically displays the Time after period of time.
  * @param  void
  * @retval void
  */
void source_timeout(void) {
	static uint32_t timestamp = 0;
	static CF_Source last_src = Time;
	CF_Source src = CF_getSourceSelect();
	if (src != last_src && src != Alt) {
		timestamp = millis();
		last_src = src;
	}
	if (millis() - timestamp > 5000 && src != Alt) {
		CF_setSourceSelect(Time);
	}
}


/**
  * @brief  Change the Alternative Source.
  * @param  void
  * @retval void
  */
void Change_AltSource(void) {
	static uint32_t timestamp = 0;
	if (millis() - timestamp > 5000) {
		timestamp = millis();
		uint8_t arr[] = { 0,0,0,0 };
		CF_getAlt(arr);
		for (int i = 0; i < sizeof(arr) / sizeof(arr[0]); i++) {
			arr[i] = (arr[i] + 1) % 10;
		}
		CF_setAlt(arr);
	}
}

/**
  * @brief  Switches the Tubes on and off randomly. Can be used to flicker the tubes when calles repeatedly.
  * @param  void
  * @retval Returns true, when the sequence is over. The user should then stop calling this function and set
  *			the preferred mask manually.
  */
bool flicker_tubes(void) {
	static uint32_t timestamp = 0;
	static uint8_t counter = 0;
	static uint8_t tube_factor[] = { random(1,5),random(1,5),random(1,5),random(1,5) };
	if (millis() - timestamp > (1) ) {
		timestamp = millis();
		counter++;
		uint8_t mask[] = { 0,0,0,0 };
		CF_getMask(mask);
		for (int i = 0; i < 4; i++) {
			if (!random(counter/tube_factor[i])) {
				mask[i] = !mask[i];
			}
		}

		uint8_t temp = mask[0];
		mask[0] = mask[2];
		mask[2] = temp;

		CF_setMask(mask);

		if (counter == 0) {
			tube_factor[0] = random(1, 5);
			tube_factor[1] = random(1, 5);
			tube_factor[2] = random(1, 5);
			tube_factor[3] = random(1, 5);
			return true;
		}
	}
	return false;
}



/**
  * @brief  Determines whether the timout after a Button Action is still running.
  * @param  void
  * @retval Returns true, when the Timeout is still running, otherwise false.
  */
bool svc_btn_timeout() {
	static uint32_t timestamp = 0;
	if (millis() - timestamp > 20) {
		timestamp = millis();
		return false;
	}
	return true;
}
