/**
  ******************************************************************************
  * @file    VisualUserInterface.cpp
  * @author  Pascal Guttmann
  * @version v1
  * @date    08 Nov 2020
  * @brief   This file contains the sourcecode for the Visual User Interface.
  ******************************************************************************
  */

#include "VisualUserInterface.h"
#include <Arduino.h>

#define VUI_LONGPRESS_DURATION 300

//BUTTON
volatile uint32_t VUI_btn_timestamp;
volatile bool VUI_btn_shortPress = false;
volatile bool VUI_btn_longPress = false;

//SERVICE BUTTON
bool VUI_svcbtn_lastState = true;


/**
  * @brief  The VUI_setup function is used to initialize the Visual user Interface.
  * @param  void
  * @retval void
  */
void VUI_setup() {
	digitalWrite(VUI_BKLGHT, 0);
	digitalWrite(VUI_SLED_R, 0);
	digitalWrite(VUI_SLED_G, 0);
	digitalWrite(VUI_SLED_B, 0);

	pinMode(VUI_BKLGHT, OUTPUT);
	pinMode(VUI_SLED_R, OUTPUT);
	pinMode(VUI_SLED_G, OUTPUT);
	pinMode(VUI_SLED_B, OUTPUT);

	pinMode(VUI_SVCBTN_PIN, INPUT_PULLUP);
	pinMode(VUI_BTN_PIN, INPUT_PULLUP);
	attachInterrupt(digitalPinToInterrupt(VUI_BTN_PIN), ISR_BUTTON_PRESS, CHANGE);

#ifdef VUI_DEBUG
	Serial.println("Visual User Interface initialized.");
#endif 
}

/**
  * @brief  Determines whether the timout after a Button Action is still running.
  * @param  void
  * @retval Returns true, when the Timeout is still running, otherwise false.
  */
bool VUI_btn_actiontimeout() {
	static uint32_t timestamp = 0;
	if (millis() - timestamp > 20) {
		timestamp = millis();
		return false;
	}
	return true;
}

/**
  * @brief  The VUI_btn_short function returns true, when the Button was pressed shortly.
  * @param  void
  * @retval Returns true, when the "Button" was released within the LONGPRESS_DURATION since being pressed.
  */
bool VUI_btn_short() {
	bool temp = VUI_btn_shortPress;
	VUI_btn_shortPress = false;
#ifdef VUI_DEBUG
	if (temp) {
		Serial.println("Short Button Press Action.");
	}
#endif
	return temp;
}

/**
  * @brief  The VUI_btn_long function returns true, when the Button was pressed long.
  * @param  void
  * @retval Returns true, when the "Button" was released after the LONGPRESS_DURATION passed since being pressed.
  */
bool VUI_btn_long() {
	bool temp = VUI_btn_longPress;
	VUI_btn_longPress = false;
#ifdef VUI_DEBUG
	if (temp) {
		Serial.println("Long Button Press Action.");
	}
#endif
	return temp;
}

/**
  * @brief  The VUI_svcbtn function returns the State of the "Service Button". The State is typedefined in enum "VUI_ButtonState".
  * @param  void
  * @retval "0" == UnTuched, "1" == Pressed, "2" == Released, "3" == Hold
  */
VUI_ButtonState VUI_svcbtn() {
	bool svcbtn = digitalRead(VUI_SVCBTN_PIN);
	uint8_t ButtonState = 0;
	if (!svcbtn) {
		ButtonState += 1;
	}
	if (!VUI_svcbtn_lastState) {
		ButtonState += 2;
	}
	VUI_svcbtn_lastState = svcbtn;

#ifdef VUI_DEBUG
	//Serial.print("Service Button State = ");
	//Serial.println((VUI_ButtonState)ButtonState);
#endif

	return (VUI_ButtonState)ButtonState;
}

/**
  * @brief  The ISR_BUTTON_PRESS Interupt Service Routine stores a timestamp on a falling edge, and sets the corresponding flag for the "Button" on a rising edge.
  * @param  void
  * @retval void
  */
void ISR_BUTTON_PRESS() {
	EIMSK &= ~(1 << INT1);

	static volatile bool lastPinState = HIGH;
	uint32_t timestamp = millis();
	bool pinState = digitalRead(VUI_BTN_PIN);

#ifdef VUI_DEBUG
	Serial.print("ISR State ==\t");
	Serial.print(pinState);
	Serial.print("\t@time:\t");
	Serial.println(timestamp);
#endif

	if (pinState == lastPinState) {
#ifdef VUI_DEBUG
		Serial.println("ISR PinState Equal... returning");
#endif
	}
	else {
		lastPinState = pinState;

		if (pinState == LOW) {
			VUI_btn_timestamp = timestamp;
		}
		else {
			if (!VUI_btn_actiontimeout()) {
				if (timestamp - VUI_btn_timestamp < VUI_LONGPRESS_DURATION) {
					VUI_btn_shortPress = true;
				}
				else {
					VUI_btn_longPress = true;
				}
			}
		}
	}

	EIMSK |= (1 << INT1);
	return;
}
